import shutil
from os import makedirs
from os.path import dirname, exists, expanduser, join

CONFIG_FILE_NAME = 'config.yml'


class ConfigFile():

    def __init__(self):
        if not exists(join(self._config_path(), CONFIG_FILE_NAME)):
            self._write_default_config()

    def file_path(self) -> str:
        return join(self._config_path(), CONFIG_FILE_NAME)

    def _config_path(self) -> str:
        return expanduser('~/.config/local_redash/')

    def _write_default_config(self) -> None:
        makedirs(self._config_path(), exist_ok=True)

        default_config = self._get_default_config()
        shutil.copyfile(default_config,
                        join(self._config_path(), CONFIG_FILE_NAME))

    def _get_default_config(self) -> str:
        from local_redash import __file__ as package_root

        package_root = dirname(package_root)
        return join(package_root, CONFIG_FILE_NAME)



import os

from local_redash.commands.base import Command, ResultData
from local_redash.lib.redash_client import RedashClient


class RunQueryCommand(Command):

    def __init__(self, client: RedashClient) -> None:
        self._redash_client = client

    def execute(self, query_id: int) -> ResultData:
        result = self._redash_client.query_result(query_id)

        if result is None:
            return []

        return self._sort_columns(result).rows.dict()



from dependency_injector import containers, providers

from local_redash.command_executer import CommandExecuter
from local_redash.commands.data_source_list import DataSourceListCommand
from local_redash.commands.export_query import ExportQueryCommand
from local_redash.commands.query import QueryCommand
from local_redash.commands.query_list import QueryListCommand
from local_redash.commands.run_query import RunQueryCommand
from local_redash.commands.show_query import ShowQueryCommand
from local_redash.lib.redash_client import RedashClient


class Container(containers.DeclarativeContainer):

    config = providers.Configuration()
    redash_client = providers.Factory(
        RedashClient,
        redash_url=config.redash.url,
        api_key=config.redash.api_key,
    )

    # Commands
    data_source_list_command = providers.Singleton(
        DataSourceListCommand,
        client=redash_client,
    )

    query_command = providers.Singleton(
        QueryCommand,
        client=redash_client,
    )

    run_query_command = providers.Singleton(
        RunQueryCommand,
        client=redash_client,
    )

    query_list_command = providers.Singleton(QueryListCommand,
                                             client=redash_client,
                                             columns=config.columns.query_list)

    export_query_command = providers.Singleton(
        ExportQueryCommand,
        client=redash_client,
    )

    show_query_command = providers.Singleton(
        ShowQueryCommand,
        client=redash_client,
    )

    # Command to execute
    command = providers.Selector(
        config.command.type,
        data_source_list=data_source_list_command,
        query=query_command,
        query_list=query_list_command,
        export_query=export_query_command,
        show_query=show_query_command,
        run_query=run_query_command,
    )

    executer = providers.Factory(
        CommandExecuter,
        command=command,
        tablefmt=config.table_format.tablefmt,
        stralign=config.table_format.stralign,
        maxcolwidths=config.table_format.maxcolwidths,
    )



import os

from local_redash.commands.base import Command, ResultData
from local_redash.lib.redash_client import RedashClient


class QueryCommand(Command):

    def __init__(self, client: RedashClient) -> None:
        self._redash_client = client

    def execute(self, query_path: str, data_source_id: str) -> ResultData:
        query_str = self._get_query(query_path)
        query_name = self._get_file_name(query_path)

        target_query = self._redash_client.search_query(query_name)

        if target_query is None:
            created_query = self._redash_client.create_query(
                data_source_id, query_name, query_str)
            result = self._redash_client.query_result(created_query.id)

        else:
            self._redash_client.update_query(target_query.id, query_str)
            result = self._redash_client.query_result(target_query.id)

        if result is None:
            return []

        return self._sort_columns(result).rows.dict()

    def _get_query(self, query_file_path: str) -> str:
        with open(query_file_path, 'r', encoding='utf-8') as f:
            query = f.read()
        return query

    def _get_file_name(self, file_path: str) -> str:
        return os.path.basename(file_path).split('.')[0]



from local_redash.commands.base import Command, ResultData
from local_redash.lib.redash_client import RedashClient


class QueryListCommand(Command):

    def __init__(self, client: RedashClient, columns: list = []) -> None:
        self._redash_client = client
        self._columns = columns

    def execute(self, sort_column: str | None = None) -> ResultData:
        query_list = self._redash_client.get_query_list()
        keys = set(self._columns)
        filtered_query_list = self.filter_columns(query_list.dict(), keys)

        if sort_column is None:
            return filtered_query_list

        return self.sort_records(filtered_query_list, sort_column)



from enum import Enum, IntEnum

from pydantic import BaseModel, Field


class User(BaseModel):
    id: int
    name: str
    auth_type: str
    email: str
    groups: list[int]
    is_disabled: bool
    is_email_verified: bool
    is_invitation_pending: bool
    profile_image_url: str
    created_at: str
    active_at: str
    updated_at: str
    disabled_at: str | None


class QuerySchedule(BaseModel):
    interval: int | None
    until: str | None
    day_of_week: str | None
    time: str | None


class Query(BaseModel):
    id: int
    name: str
    description: str | None
    api_key: str
    is_archived: bool
    is_draft: bool
    is_favorite: bool
    is_safe: bool
    data_source_id: int
    last_modified_by_id: int
    latest_query_data_id: int | None
    options: dict
    query: str
    query_hash: str
    runtime: float | None
    schedule: QuerySchedule | None
    tags: list[str]
    version: int
    user: User
    retrieved_at: str | None
    updated_at: str
    created_at: str


class QueryList(BaseModel):
    __root__: list[Query]

    def __iter__(self):
        return iter(self.__root__)

    def __getitem__(self, item):
        return self.__root__[item]

    def dict(self, **kwargs):
        return super().dict(**kwargs)['__root__']


class ResponseQuery(BaseModel):
    count: int
    page: int
    page_size: int
    results: list[Query]


class LastModified(BaseModel):
    id: int
    name: str
    email: str
    auth_type: str
    is_disabled: bool
    profile_image_url: str
    is_invitation_pending: bool
    groups: list[int]
    is_email_verified: bool
    created_at: str
    active_at: str
    updated_at: str
    disabled_at: str | None


class Visualization(BaseModel):
    id: int
    name: str
    description: str
    type: str
    options: dict
    created_at: str
    updated_at: str


class QueryDetail(BaseModel):
    id: int
    name: str
    api_key: str
    description: str | None
    is_archived: bool
    is_draft: bool
    is_favorite: bool
    is_safe: bool
    data_source_id: int
    last_modified_by: LastModified
    latest_query_data_id: int | None
    options: dict
    query: str
    query_hash: str
    schedule: QuerySchedule | None
    tags: list[str]
    version: int
    user: User
    visualizations: list[Visualization]
    updated_at: str
    created_at: str


class DataSourceType(Enum):
    POSTGRESQL = 'pg'
    MYSQL = 'mysql'
    BIGQUERY = 'bigquery'
    PYTHON = 'python'
    ATHENA = 'athena'
    GOOGLE_SPREADSHEETS = 'google_spreadsheets'
    RESULTS = 'results'


class DataSource(BaseModel):
    id: int
    name: str
    pause_reason: str | None
    syntax: str
    paused: int
    view_only: bool
    type: DataSourceType

    class Config:
        use_enum_values = True


class DataSourceDetail(BaseModel):
    id: int
    name: str
    scheduled_queue_name: str
    pause_reason: str | None
    queue_name: str
    syntax: str
    paused: int
    type: DataSourceType
    groups: dict[str, str | int | bool]
    options: dict[str, str | int | bool]


class DataSourceList(BaseModel):
    __root__: list[DataSource]

    def __iter__(self):
        return iter(self.__root__)

    def __getitem__(self, item):
        return self.__root__[item]

    def dict(self, **kwargs):
        return super().dict(**kwargs)['__root__']


class QueryResulColumn(BaseModel):
    friendly_name: str
    type: str | None
    name: str


class QueryResulRows(BaseModel):
    __root__: list[dict[str, str | int | float | bool | None]]

    def __iter__(self):
        return iter(self.__root__)

    def __getitem__(self, item):
        return self.__root__[item]

    def dict(self, **kwargs):
        return super().dict(**kwargs)['__root__']


class QueryResultData(BaseModel):
    rows: QueryResulRows
    columns: list[QueryResulColumn]


class QueryResult(BaseModel):
    id: int
    data_source_id: int
    query_hash: str
    query: str
    runtime: float
    data: QueryResultData
    retrieved_at: str


class JobResultStatus(IntEnum):
    QUEUED = 1
    STARTED = 2
    FINISHED = 3
    FAILED = 4


class JobResult(BaseModel):
    id: str
    status: JobResultStatus
    query_result_id: int | None
    error: str
    updated_at: int

    class Config:
        use_enum_values = True


class SqlFormatDialects():

    DATASOURCE_DIALECTS_MMAPPING = {
        DataSourceType.ATHENA: 'athena',
        DataSourceType.BIGQUERY: 'bigquery',
        DataSourceType.MYSQL: 'mysql',
        DataSourceType.POSTGRESQL: 'postgres'
    }

    @classmethod
    def from_datasource_type(cls, source_type: DataSourceType) -> str:

        if source_type in cls.DATASOURCE_DIALECTS_MMAPPING:
            return cls.DATASOURCE_DIALECTS_MMAPPING[source_type]
        else:
            return 'ansi'



from local_redash.commands.base import Command, ResultData
from local_redash.lib.redash_client import RedashClient


class DataSourceListCommand(Command):

    def __init__(self, client: RedashClient) -> None:
        self._redash_client = client

    def execute(self, sort_column: str | None = None) -> ResultData:
        data_source_list = self._redash_client.get_data_source_list()

        if sort_column is None:
            return data_source_list.dict()

        return self.sort_records(data_source_list.dict(), sort_column)



import json
from typing import Callable, TypeAlias, Union

import sqlfluff
from black import FileMode, format_file_contents
from local_redash.models.redash_client import (DataSourceType, QueryResulRows,
                                               QueryResultData,
                                               SqlFormatDialects)

ResultDataRow: TypeAlias = dict[str, Union[str, int, dict]]
ResultData: TypeAlias = list[ResultDataRow]


class Command:

    def __init__(self) -> None:
        pass

    def execute(self, *args) -> ResultData:
        raise NotImplementedError()

    def filter_columns(self, result_data: ResultData,
                       columns: set[str]) -> ResultData:

        check: Callable[[tuple[str, str | int]],
                        bool] = lambda item: item[0] in columns
        return list(
            map(lambda row: dict(filter(check, row.items())), result_data))

    def sort_records(self, result_data: ResultData,
                     column_name: str) -> ResultData:
        return sorted(result_data,
                      key=lambda row: self._sort_key(row[column_name]))

    def format_query(self, query_str: str,
                     data_source_type: DataSourceType) -> str:

        if data_source_type == DataSourceType.PYTHON:
            return self.format_python_code(query_str)
        else:
            return self.format_sql(query_str, data_source_type)

    def format_sql(self, query_str: str,
                   data_source_type: DataSourceType) -> str:
        dialect = SqlFormatDialects.from_datasource_type(data_source_type)

        return sqlfluff.fix(
            query_str,
            dialect=dialect,
        )

    def format_python_code(self, python_code: str) -> str:
        return format_file_contents(python_code, fast=False, mode=FileMode())

    def _sort_key(self, value: str | int | dict) -> str | int:
        if isinstance(value, int):
            return value
        else:
            return json.dumps(value)

    def _sort_columns(self, query_result_data: QueryResultData):
        """
        To be in the order of sql SELECT.
        Sort in the same order as the columns key.
        """
        columns = [
            resul_column.friendly_name
            for resul_column in query_result_data.columns
        ]

        sorted_result_data_dict_list = [{
            key: row[key]
            for key in columns
        } for row in query_result_data.rows.dict()]

        return QueryResultData(
            rows=QueryResulRows.parse_obj(sorted_result_data_dict_list),
            columns=query_result_data.columns)



from local_redash.commands.base import Command, ResultData
from local_redash.lib.redash_client import RedashClient


class ShowQueryCommand(Command):

    def __init__(self, client: RedashClient) -> None:
        self._redash_client = client

    def execute(self, query_id: int) -> ResultData:
        query = self._redash_client.get_query(query_id)

        if query is not None:
            data_source = self._redash_client.get_data_source(
                query.data_source_id)
            formatted_query = self.format_query(query.query, data_source.type)
            return [{'query': formatted_query}]
        else:
            return []



import json
import sys
import time
from ctypes import resize
from typing import Any

import httpx
from local_redash.exceptions import AuthenticationError
from local_redash.models.redash_client import (DataSourceDetail,
                                               DataSourceList, JobResult,
                                               JobResultStatus, Query,
                                               QueryDetail, QueryList,
                                               QueryResultData)
from timeout_decorator import timeout

QUERY_TIME_OUT = 10


class RedashClient:

    def __init__(self, redash_url: str, api_key: str) -> None:
        if redash_url.endswith('/'):
            redash_url = redash_url[:-1]

        self.redash_url = redash_url
        self.request_headers = {"Authorization": f"Key {api_key}"}
        self._authentication()

    def search_query(self, search_name: str) -> Query | None:
        all_queries = self.get_query_list()
        result = None
        for query in all_queries:
            if query.name == search_name:
                result = query
                break

        return result

    def update_query(self,
                     query_id: int,
                     query: str,
                     options: dict | None = None) -> QueryDetail:

        if options is None or not isinstance(options, dict):
            options = {}

        payload = {'query': query, 'options': options}
        result = self._post(f'api/queries/{query_id}', payload)
        return QueryDetail.parse_obj(result)

    def create_query(self,
                     data_source_id: str,
                     name: str,
                     query: str,
                     description: str = "",
                     options: dict | None = None) -> QueryDetail:

        if options is None or not isinstance(options, dict):
            options = {}

        try:
            payload = {
                "data_source_id": data_source_id,
                "name": name,
                "query": query,
                "description": description,
                "options": options
            }
            result = self._post('api/queries', payload)
            return QueryDetail.parse_obj(result)
        except httpx.HTTPStatusError as exc:
            # Without printing a traceback
            sys.tracebacklimit = -1
            raise Exception('Failed to create query.') from None

    def get_query(self, id: int) -> QueryDetail | None:
        try:
            response = self._get(f'api/queries/{id}')
            return QueryDetail.parse_obj(response)
        except httpx.HTTPStatusError as exc:
            if exc.response.status_code == 404:
                return None
            else:
                raise exc

    def get_data_source(self, id: int) -> DataSourceDetail:
        response = self._get(f'api/data_sources/{id}')
        return DataSourceDetail.parse_obj(response)

    def get_data_source_list(self) -> DataSourceList:
        response = self._get('api/data_sources')
        return DataSourceList.parse_obj(response)

    def get_query_list(self) -> QueryList:
        result = self._get_paginate('api/queries')
        return QueryList.parse_obj(result)

    def query_result(self, query_id: int, params={}) -> QueryResultData | None:
        try:
            payload = {'max_age': 0, 'parameters': params}
            results_response = self._post(f'api/queries/{query_id}/results',
                                          payload)
        except httpx.HTTPStatusError as exc:
            if exc.response.status_code == 404:
                return None
            else:
                raise exc

        result_id = self._polling_job(results_response['job']['id'])

        if result_id is None:
            raise Exception('Query execution timed out.')

        response = self._get(
            f'api/queries/{query_id}/results/{result_id}.json')

        return QueryResultData.parse_obj(response['query_result']['data'])

    def _authentication(self):
        # Request for authentication check. Because each API has a
        # different response code for authentication errors.
        self._get('api/users')

    @timeout(QUERY_TIME_OUT)
    def _polling_job(self, job_id: str) -> int | None:
        job_status: int | None = None
        job_query_result_id: int | None = None

        while True:
            response = self._get(f'api/jobs/{job_id}')

            job_result = JobResult.parse_obj(response['job'])
            job_status = job_result.status
            if job_status == JobResultStatus.FINISHED:
                job_query_result_id = job_result.query_result_id
                break
            if job_status == JobResultStatus.FAILED:
                # Without printing a traceback
                sys.tracebacklimit = -1
                print(job_result.error)
                raise Exception('Query execution failed.')

            time.sleep(1)

        return job_query_result_id

    def _get_paginate(self, path: str, params={}) -> list:
        if not 'page' in params:
            params = {**params, **{'page': 1}}

        response = self._get(path, params)
        results = response['results']
        page = response['page']
        page_size = response['page_size']

        if page * page_size >= response['count']:
            return results
        else:
            return [
                *results,
                *self._get_paginate(path, {
                    **params,
                    **{
                        'page': page + 1
                    }
                }),
            ]

    def _get(self, path: str, params=None) -> Any:
        response = None
        try:
            response = httpx.get(f"{self.redash_url}/{path}",
                                 headers=self.request_headers,
                                 params=params)
            response.raise_for_status()
            return response.json()
        except Exception as exc:
            if self._is_authentication_error(response):
                # Without printing a traceback
                sys.tracebacklimit = -1
                raise AuthenticationError() from None
            else:
                self._print_response_error(response)
                raise exc

    def _post(self, path: str, payload=None) -> Any:
        response = None
        try:
            response = httpx.post(f"{self.redash_url}/{path}",
                                  headers=self.request_headers,
                                  json=payload)
            response.raise_for_status()
            return response.json()
        except Exception as exc:
            if self._is_authentication_error(response):
                # Without printing a traceback
                sys.tracebacklimit = -1
                raise AuthenticationError() from None
            else:
                self._print_response_error(response)
                raise exc

    def _print_response_error(self, response: httpx.Response | None) -> None:
        print('Error: API request failed.')
        if response is not None:
            if self._is_json(response.content):
                print(f'Response body: {response.json()}')
            else:
                print(f'Response body: {response.text}')

    def _is_json(self, json_constant) -> bool:
        try:
            json.loads(json_constant)
        except ValueError as e:
            return False
        return True

    def _is_authentication_error(self,
                                 response: httpx.Response | None) -> bool:
        if response is not None and self._is_json(
                response.content) and 'message' in response.json().keys():
            if response.json(
            )['message'] == "Couldn't find resource. Please login and try again.":
                return True

        return False



import json

from tabulate import tabulate

from local_redash.commands.base import Command, ResultDataRow


class CommandExecuter:

    def __init__(self, command: Command, tablefmt: str, stralign: str,
                 maxcolwidths: int):
        self._command = command
        self._tablefmt = tablefmt
        self._stralign = stralign
        self._maxcolwidths = maxcolwidths

    def execute(
        self,
        *args,
        headers: str = 'keys',
        tablefmt: str | None = None,
        stralign: str | None = None,
        maxcolwidths: int | None = None,
    ) -> None:
        tablefmt_val = tablefmt or self._tablefmt
        stralign_val = stralign or self._stralign
        maxcolwidths_val = maxcolwidths or self._maxcolwidths

        result = self._command.execute(*args)
        converted_result = list(
            map(lambda row: self.convert_values_to_string(row), result))

        if len(converted_result) == 0:
            print('Data does not exist.')
            return

        print(
            tabulate(
                converted_result,
                headers=headers,
                tablefmt=tablefmt_val,
                stralign=stralign_val,
                maxcolwidths=maxcolwidths_val,
            ))

    def convert_values_to_string(self, row: ResultDataRow) -> ResultDataRow:
        for key, value in row.items():
            if isinstance(value, dict):
                row[key] = json.dumps(value, ensure_ascii=False)
            else:
                row[key] = str(value)
        return row



import os
from os.path import dirname, join

import click
from click_option_group import RequiredMutuallyExclusiveOptionGroup, optgroup
from dependency_injector.wiring import Provide, inject
from dotenv import load_dotenv

from local_redash.containers import Container
from local_redash.lib.config_file import ConfigFile

os.environ['NO_PROXY'] = '127.0.0.1,localhost'


@click.group()
@click.pass_context
@inject
def main(ctx, container: Container = Provide[Container]):
    container.config.command.type.from_value(
        ctx.invoked_subcommand.replace('-', '_'))
    command_executer = container.executer()
    ctx.obj = command_executer


@main.command()
@click.option('--query-file', required=True, type=str, help='')
@click.option('--data-source-id', required=True, type=int, help='')
@click.pass_context
def query(ctx, query_file, data_source_id):
    click.echo('Run query in Sql file.')
    ctx.obj.execute(query_file, data_source_id)


@main.command()
@click.option('--query-id', required=True, type=int, help='')
@click.pass_context
def run_query(ctx, query_id):
    click.echo('Run query.')
    ctx.obj.execute(query_id)


@main.command()
@click.option('--sort-column',
              type=str,
              default='id',
              show_default=True,
              help='')
@click.pass_context
def data_source_list(ctx, sort_column):
    click.echo('List of Data Sources')
    ctx.obj.execute(sort_column)


@main.command()
@click.option('--sort-column',
              type=str,
              default='id',
              show_default=True,
              help='')
@click.pass_context
def query_list(ctx, sort_column):
    click.echo('List of Query')
    ctx.obj.execute(sort_column)


@main.command()
@optgroup.group('Specify query',
                cls=RequiredMutuallyExclusiveOptionGroup,
                help='Specify query to export')
@optgroup.option('--query-name', type=str, help='Query name')
@optgroup.option('--query-id', type=int, help='Query id')
@click.option('--file-path', required=True, type=str, help='')
@click.pass_context
def export_query(ctx, query_name, query_id, file_path):
    click.echo('Export Query')
    query_key: str | int = query_name if query_name else query_id
    ctx.obj.execute(query_key, file_path, stralign='left')


@main.command()
@click.option('--query-id', required=True, type=int, help='')
@click.pass_context
def show_query(ctx, query_id):
    click.echo('Show Query')
    ctx.obj.execute(query_id)


def initialize():
    load_dotenv(join(dirname(__file__), '.env'))

    container = Container()
    # redash-api
    container.config.redash.url.from_env("REDASH_URL")
    container.config.redash.api_key.from_env("API_KEY")
    # config
    config_file = ConfigFile()
    container.config.from_yaml(config_file.file_path())

    container.wire(modules=[__name__])

    main()


if __name__ == '__main__':
    initialize()



import os

from local_redash.commands.base import Command, ResultData
from local_redash.lib.redash_client import RedashClient
from local_redash.models.redash_client import QueryDetail


class ExportQueryCommand(Command):

    def __init__(self, client: RedashClient) -> None:
        self._redash_client = client

    def execute(self, query_key: str | int, file_path: str) -> ResultData:
        if not os.path.isdir(file_path):
            return []
        file_path = file_path.rstrip(os.path.sep)

        target_query: QueryDetail | None = None

        if type(query_key) == str:
            target_query = self.from_query_name(query_key)
        elif type(query_key) == int:
            target_query = self.from_query_id(query_key)
        else:
            pass

        if target_query is None:
            return []

        data_source = self._redash_client.get_data_source(
            target_query.data_source_id)

        formatted_query = self.format_query(target_query.query,
                                            data_source.type)
        result = self._save_query(formatted_query,
                                  f'{file_path}/{target_query.name}.sql')

        if not result:
            return []

        return [{'exported-query': formatted_query}]

    def from_query_id(self, query_id: int) -> QueryDetail | None:
        return self._redash_client.get_query(query_id)

    def from_query_name(self, query_name: str) -> QueryDetail | None:
        query = self._redash_client.search_query(query_name)

        if query is None:
            return None

        return self._redash_client.get_query(query.id)

    def _save_query(self, query_str: str, file_path: str) -> bool:
        try:
            with open(file_path, 'w', encoding='utf-8') as f:
                f.write(query_str)
        except IOError:
            return False

        return True
